﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Windows.Forms;

namespace LandformsClassifier
{   
    /// <summary>
    /// Represents a landscape loaded from a file with a concrete point set.
    /// </summary>
    class Terrain
    {
        
        private const int MAX_DIST = 10; // maximum number of cells for finding the nearest point in a specific direction 
        private const double TO_DEGREES = 180.0 / Math.PI; // constant for converting radians to degrees
        private const int SQUARE_SIZE = 5; // size of the square area
        private int POINT_SIZE = 1; // size (diameter) of the point

        public bool isActiveZoom = false; // if we are in the zoom
        public bool availableZoom = false; // if we can zoom the terrain
        private int row, column; // row and column of the left upper corner of the zoomed area
        private bool[] pars = new bool[5]; // true, if the homogeneity is computed based on particular parameter (0=z, 1=slope, 2=aspect, 3=prof.cur., 4=plan cur.)
        private bool[] terFeatures = new bool[5]; // true = draw particular terrain feature, (0=flat, 1=ridges/valleys, 2=terrain edges, 3=uniform slope, 4=curved slope)
        private Graphics g; // graphics context for drawing on the canvas
        private Color bgColor; // background color of the canvas
        private PictureBox canvas;
        private Color[,] imageBuff; // buffer to store colors of the points
        private Point[,] grid; // 2D array representing a terrain
        private Area[,] areas; // grid divided into larger areas
        private Reader rd = new Reader(); // reader for the loading points
        private float ratioX, ratioY;
        private float minHz = 9999, minHs = 9999, minHa = 9999, minHpr = 9999, minHpl = 9999;
        private float maxHz = 0, maxHs = 0, maxHa = 0, maxHpr = 0, maxHpl = 0;
        private float omegaHz, omegaHs, omegaHa, omegaHpr, omegaHpl;

        List<Area> lowestAreas = new List<Area>(); // the areas with the lowest average z in comparison with all neighbours
        List<Area> highestAreas = new List<Area>(); // the areas with the highest average z in comparison with all neighbours

        public Terrain(Graphics g, PictureBox canvas)
        {
            this.g = g;
            this.canvas = canvas;
            bgColor = Color.Black;
            g.Clear(Color.White);
            canvas.Invalidate();
        }

        /** Load and process all loaded points from the file */
        public void LoadFile(String filename)
        {
            g.Clear(Color.White);
            canvas.Invalidate();
            grid = rd.LoadPoints(filename);
            if (grid == null) return;
            if (rd.rangeX > canvas.Width || rd.rangeY > canvas.Height)
                POINT_SIZE = 1;
            else
                POINT_SIZE = 2;

            imageBuff = new Color[grid.GetLength(0), grid.GetLength(1)];

            ProcessGrid();
            ClassifyPoints();
            CreateSquareAreas();
            AnalyzeSquareAreas();
        }

        /** Go through all points except border values and compute parameters.
            Then go again and compute homogeneity indeces and scale their values 
            into the range <0 - 1> */
        private void ProcessGrid()
        {
            int i, j;
            int maxI = grid.GetLength(0) - 1;
            int maxJ = grid.GetLength(1) - 1;

            for (i = 1; i < maxI; i++)
            {
                for (j = 1; j < maxJ; j++)
                {
                    if (grid[i, j] != null)
                        ComputeParameters(i, j);
                }
            }

            maxI = grid.GetLength(0) - 2;
            maxJ = grid.GetLength(1) - 2;

            for (i = 2; i < maxI; i++)
            {
                for (j = 2; j < maxJ; j++)
                {
                    if (grid[i, j] != null)
                    {
                        Point[] window = GetWindow(i, j, false);
                        ComputeHindices(window);
                    }
                }
            }

            // compute ranges of the H indeces
            omegaHz = maxHz - minHz;
            omegaHs = maxHs - minHs;
            omegaHa = maxHa - minHa;
            omegaHpr = maxHpr - minHpr;
            omegaHpl = maxHpl - minHpl;

            // scale all H values into interval <0,1>
            for (i = 2; i < maxI; i++)
            {
                for (j = 2; j < maxJ; j++)
                {
                    if (grid[i, j] != null)
                    {
                        grid[i, j].Hz /= omegaHz;
                        grid[i, j].Hs /= omegaHs;
                        grid[i, j].Ha /= omegaHa;
                        grid[i, j].Hpr /= omegaHpr;
                        grid[i, j].Hpl /= omegaHpl;
                    }
                }
            }
        }

        /** Compute homogeneity indeces */
        private void ComputeHindices(Point[] window)
        {
            float con = 1 / (float) Math.Sqrt(2); // constant value
            float unit_vx = 0, unit_vy = 0; // x and y values of unit separation vector
            float sumGxiZ = 0, sumGyiZ = 0; // sum of x and y values of gradient vector of z coordinate
            float sumGxiSlope = 0, sumGyiSlope = 0; // sum of x and y values of gradient vector of slope
            float sumGxiAspect = 0, sumGyiAspect = 0; // sum of x and y values of gradient vector of aspect
            float sumGxiProfileCur = 0, sumGyiProfileCur = 0; // sum of x and y values of gradient vector of profile curvature
            float sumGxiPlanCur = 0, sumGyiPlanCur = 0; // sum of x and y values of gradient vector of plan curvature
            float temp;

            float Hz, Hs, Ha, Hpr, Hpl;

            for (int i = 0; i < window.Length; i++)
            {
                if (window[i] == window[4])
                    continue;
                // set unit vector values
                switch (i)
                {
                    case 0:
                        unit_vx = -con;
                        unit_vy = -con;
                        break;
                    case 1:
                        unit_vx = 0;
                        unit_vy = -1;
                        break;
                    case 2:
                        unit_vx = con;
                        unit_vy = -con;
                        break;
                    case 3:
                        unit_vx = -1;
                        unit_vy = 0;
                        break;
                    case 5:
                        unit_vx = 1;
                        unit_vy = 0;
                        break;
                    case 6:
                        unit_vx = -con;
                        unit_vy = con;
                        break;
                    case 7:
                        unit_vx = 0;
                        unit_vy = 1;
                        break;
                    case 8:
                        unit_vx = con;
                        unit_vy = con;
                        break;
                }
   
                // homogeneity for Z-coordinate
                temp = window[i].z - window[4].z;
                sumGxiZ += (temp * unit_vx);
                sumGyiZ += (temp * unit_vy);

                // homogeneity for slope
                temp = window[i].slope - window[4].slope;
                sumGxiSlope += (temp * unit_vx);
                sumGyiSlope += (temp * unit_vy);

                // homogeneity for aspect
                temp = Math.Abs(window[i].aspectDeg - window[4].aspectDeg);
                if (temp > 180)
                    temp = (360 - temp);
                sumGxiAspect += (temp * unit_vx);
                sumGyiAspect += (temp * unit_vy);

                // homogeneity for profile curvature
                temp = Math.Abs(window[i].profileCurvature - window[4].profileCurvature);
                sumGxiProfileCur += (temp * unit_vx);
                sumGyiProfileCur += (temp * unit_vy);
                
                // homogeneity for plan curvature
                temp = Math.Abs(window[i].planCurvature - window[4].planCurvature);
                sumGxiPlanCur += (temp * unit_vx);
                sumGyiPlanCur += (temp * unit_vy);
            }

            Hz = (float) Math.Sqrt(sumGxiZ * sumGxiZ + sumGyiZ * sumGyiZ);
            Hs = (float) Math.Sqrt(sumGxiSlope * sumGxiSlope + sumGyiSlope * sumGyiSlope);
            Ha = (float) Math.Sqrt(sumGxiAspect * sumGxiAspect + sumGyiAspect * sumGyiAspect);
            Hpr = (float) Math.Sqrt(sumGxiProfileCur * sumGxiProfileCur + sumGyiProfileCur * sumGyiProfileCur);
            Hpl = (float) Math.Sqrt(sumGxiPlanCur * sumGxiPlanCur + sumGyiPlanCur * sumGyiPlanCur);
            
            window[4].Hz = Hz;
            window[4].Hs = Hs;
            window[4].Ha = Ha;
            window[4].Hpr = Hpr;
            window[4].Hpl = Hpl;

            if (Hz < minHz)
                minHz = Hz;
            if (Hz > maxHz)
                maxHz = Hz;

            if (Hs < minHs)
                minHs = Hs;
            if (Hs > maxHs)
                maxHs = Hs;
            
            if (Ha < minHa)
                minHa = Ha;
            if (Ha > maxHa)
                maxHa = Ha;
                
            if (Hpr < minHpr)
                minHpr = Hpr;
            if (Hpr > maxHpr)
                maxHpr = Hpr;
                
            if (Hpl < minHpl)
                minHpl = Hpl;
            if (Hpl > maxHpl)
                maxHpl = Hpl;
        }

        /** Compute overall index homogenity using all chosen parameters */
        public void ComputeOverallH(double Tz, double Ts, double Ta, double Tpr, double Tpl)
        {
            float tmp, tmpZ, tmpSlope, tmpAspect, tmpProfileCur, tmpPlanCur;
            int maxI = grid.GetLength(0) - 2;
            int maxJ = grid.GetLength(1) - 2;

            for (int i = 2; i < maxI; i++)
            {
                for (int j = 2; j < maxJ; j++)
                {
                    if (grid[i, j] == null)
                        continue;

                    // use tresholds
                    tmpZ = grid[i, j].Hz;
                    if (tmpZ < Tz)
                        tmpZ = 0;
                    
                    tmpSlope = grid[i, j].Hs;
                    if (tmpSlope < Ts)
                         tmpSlope = 0;
                    
                    tmpAspect = grid[i, j].Ha;
                    if (tmpAspect < Ta)
                         tmpAspect = 0;
                   
                    tmpProfileCur = grid[i, j].Hpr;
                    if (tmpProfileCur < Tpr)
                        tmpProfileCur = 0;

                    tmpPlanCur = grid[i, j].Hpl;
                    if (tmpPlanCur < Tpl)
                        tmpPlanCur = 0;


                    tmp = 0;
                    if (pars[0])
                        tmp += (tmpZ * tmpZ);
                    if (pars[1])
                        tmp += (tmpSlope * tmpSlope);
                    if (pars[2])
                        tmp += (tmpAspect * tmpAspect);
                    if (pars[3])
                        tmp += (tmpProfileCur * tmpProfileCur);
                    if (pars[4])
                        tmp += (tmpPlanCur * tmpPlanCur);
       
                    grid[i, j].homogeneity = (float)Math.Sqrt(tmp);
                }
            }
        }

        /// <summary>
        /// Compute parameters for the specific point.
        /// </summary>
        /// <param name="i">Row of the point in the array</param>
        /// <param name="j">Column of the point in the array</param>
        private void ComputeParameters(int i , int j)
        {
            Point[] window = GetWindow(i, j, true);

            ComputeSlope(window);
            ComputeAspect(window);
            ComputeCurvature(window);
        }

        /** Return the array of points representing window 3x3. 
         *  Empty cells are filled by the nearest point in the specific direction.
         *  param ... true, if finding nearest points is during computing parameters and not H indeces */
        private Point[] GetWindow(int i, int j, bool param)
        {
            /*  Window 3x3 stored in one dimensional array 
             *  0 1 2
             *  3 4 5
             *  6 7 8
             */
            Point[] window = new Point[9];
            window[4] = grid[i, j]; // focal point of the window

            // Specify the searching distances in each direction
            int upperDist = MAX_DIST;
            int bottomDist = MAX_DIST;
            int leftDist = MAX_DIST;
            int rightDist = MAX_DIST;
            if (param)
            {
                if (i <= MAX_DIST)
                {
                    // Upper border of the array
                    upperDist = i - 1;
                }
                else
                {
                    // Bottom border of the array
                    if (i >= (grid.GetLength(0) - 1 - MAX_DIST))
                    {
                        bottomDist = grid.GetLength(0) - 2 - i;
                    }
                }

                if (j <= MAX_DIST)
                {
                    // Left border of the array
                    leftDist = j - 1;
                }
                else
                {
                    // Right border of the array
                    if (j >= (grid.GetLength(1) - 1 - MAX_DIST))
                    {
                        rightDist = grid.GetLength(1) - 2 - j;
                    }
                }
            }
            else
            {
                if (i <= (MAX_DIST + 1))
                {
                    // Upper border of the array
                    upperDist = i - 2;
                }
                else
                {
                    // Bottom border of the array
                    if (i >= (grid.GetLength(0) - 2 - MAX_DIST))
                    {
                        bottomDist = grid.GetLength(0) - 3 - i;
                    }
                }

                if (j <= (MAX_DIST + 1))
                {
                    // Left border of the array
                    leftDist = j - 2;
                }
                else
                {
                    // Right border of the array
                    if (j >= (grid.GetLength(1) - 2 - MAX_DIST))
                    {
                        rightDist = grid.GetLength(1) - 3 - j;
                    }
                }
            }

            //           Finding of the nearest neighbours for the missing cells
            // Northwest
            if (grid[i - 1, j - 1] == null)
                window[0] = NearestNeighbourNW(i - 1, j - 1, Math.Min(upperDist, leftDist));
            else
                window[0] = grid[i - 1, j - 1];
            // North
            if (grid[i - 1, j] == null)
                window[1] = NearestNeighbourN(i - 1, j, upperDist);
            else
                window[1] = grid[i - 1, j];
            // Northeast
            if (grid[i - 1, j + 1] == null)
                window[2] = NearestNeighbourNE(i - 1, j + 1, Math.Min(upperDist, rightDist));
            else
                window[2] = grid[i - 1, j + 1];
            // East
            if (grid[i, j + 1] == null)
                window[5] = NearestNeighbourE(i, j + 1, rightDist);
            else
                window[5] = grid[i, j + 1];
            // Southeast
            if (grid[i + 1, j + 1] == null)
                window[8] = NearestNeighbourSE(i + 1, j + 1, Math.Min(bottomDist, rightDist));
            else
                window[8] = grid[i + 1, j + 1];
            // South
            if (grid[i + 1, j] == null)
                window[7] = NearestNeighbourS(i + 1, j, bottomDist);
            else
                window[7] = grid[i + 1, j];
            // Southwest
            if (grid[i + 1, j - 1] == null)
                window[6] = NearestNeighbourSW(i + 1, j - 1, Math.Min(bottomDist, leftDist));
            else
                window[6] = grid[i + 1, j - 1];
            // West
            if (grid[i, j - 1] == null)
                window[3] = NearestNeighbourW(i, j - 1, leftDist);
            else
                window[3] = grid[i, j - 1];

            return window;
        }

        
        /*
         * Compute SLOPE
         * 
         *    Window 3x3            One dimensional array
         *      a b c
         *      d e f       ->      a b c d e f g h i
         *      g h i               0 1 2 3 4 5 6 7 8
         * 
         *      slope = (arctg(Sqrt((dz / dx)^2 + (dz / dy)^2))) * 180 / PI [°]
         *      
         *      dz/dx = ((c + 2f + i) - (a + 2d + g)) / 8 * cellSizeX       , cellSizeX = 1
         *      dz/dy = ((g + 2h + i) - (a + 2b + c)) / 8 * cellSizeY       , cellSiZeY = 1
         */
        private void ComputeSlope(Point[] window)
        {
            double diffX = (((window[2].z + 2 * window[5].z + window[8].z) - (window[0].z + 2 * window[3].z + window[6].z)) / 8.0f);
            double diffY = (((window[6].z + 2 * window[7].z + window[8].z) - (window[0].z + 2 * window[1].z + window[2].z)) / 8.0f);

            window[4].slope = (float)(Math.Atan(Math.Sqrt(diffX * diffX + diffY * diffY)) * TO_DEGREES);
        }


        /*
         * Compute ASPECT
         * 
         *    Window 3x3            One dimensional array
         *      a b c
         *      d e f       ->      a b c d e f g h i
         *      g h i               0 1 2 3 4 5 6 7 8
         * 
         *      aspect = arctg2((dz / dy), -(dz / dx)) * 180 / PI [°]
         *      
         *      dz/dx = ((c + 2f + i) - (a + 2d + g)) / 8 * cellSizeX       , cellSizeX = 1
         *      dz/dy = ((g + 2h + i) - (a + 2b + c)) / 8 * cellSizeY       , cellSiZeY = 1
         */
        private void ComputeAspect(Point[] window)
        {
            double diffX = (((window[2].z + 2 * window[5].z + window[8].z) - (window[0].z + 2 * window[3].z + window[6].z)) / 8.0f);
            double diffY = (((window[6].z + 2 * window[7].z + window[8].z) - (window[0].z + 2 * window[1].z + window[2].z)) / 8.0f);

            float aspect = (float)(Math.Atan2(diffY, -diffX) * TO_DEGREES);

            // Convert into degrees in range 0 - 360°
            if (aspect > 90)
                window[4].aspectDeg = 450 - aspect;
            else
                window[4].aspectDeg = 90 - aspect;

          window[4].aspectDir = Direction.GetDirection(window[4].aspectDeg);
      //    window[4].aspectDir = Direction.GetDirection2(window[4].aspectDeg);

        }

        /******************************************************************
         *     Curvature is described as fourth-order polynomial
         *  Z = ax²y² + bx²y + cxy² + dx² + ey² + fxy + gx + hy + i
         * 
         *                 Window 3x3             One dimensional array
         *                  z1 z2 z3     ->      z1 z2 z3 z4 z5 z6 z7 z8 z9
         *                  z4 z5 z6             0  1  2  3  4  5  6  7  8
         *                  z7 z8 z9
         *                  
         *      a = [(z1 + z3 + z7 + z9) / 4  - (z2 + z4 + z6 + z8) / 2 + z5] / cellSize^4
         *      b = [(z1 + z3 - z7 - z9) / 4 - (z2 - z8) / 2] / cellSize^3
         *      c = [(-z1 + z3 - z7 + z9) / 4 + (z4 - z6)] / 2] / cellSize^3
         *      d = [(z4 + z6) / 2 - z5] / cellSize^2
         *      e = [(z2 + z8) / 2 - z5] / cellSize^2
         *      f = (-z1 + z3 + z7 - z9) / 4 * cellSize^2
         *      g = (-z4 + z6) / 2 * cellSize
         *      h = (z2 - z8) / 2 * cellSize
         *      i = z5
         *      
         *      cellSize = 1
         ******************************************************************/

        /*
         * Compute PROFILE CURVATURE
         *      Kpr = 200 * (dg^2 + eh^2 + fgh) / (g^2 + h^2)
         * Compute PLAN CURVATURE     
         *      Kpl = -200 * (dh^2 + eg^2 - fgh) / (g^2 + h^2)
         * 
         */
        private void ComputeCurvature(Point[] window)
        {
            float d = (window[3].z + window[5].z) / 2.0f - window[4].z;
            float e = (window[1].z + window[7].z) / 2.0f - window[4].z;
            float f = (-window[0].z + window[2].z + window[6].z - window[8].z) / 4.0f;
            float g = (-window[3].z + window[5].z) / 2.0f;
            float h = (window[1].z - window[7].z) / 2.0f;

            if (g == 0 && h == 0)
            {
                window[4].profileCurvature = 0;
                window[4].planCurvature = 0;
            }
            else
            {
                window[4].profileCurvature = (200 * (d * (g * g) + e * (h * h) + f * g * h)) / ((g * g) + (h * h));
                window[4].planCurvature = (-200 * (d * (h * h) + e * (g * g) - f * g * h)) / ((g * g) + (h * h));
            }
        }

        /** Find the nearest point in the NW direction */
        private Point NearestNeighbourNW(int i, int j, int dist)
        {
            int r = i - 1;
            int c = j - 1;

            for (int k = 0; k < dist; k++)
            {
                if (grid[r, c] != null)
                    return grid[r, c];
                if (grid[r + 1, c] != null)
                    return grid[r + 1, c];
                if (grid[r, c + 1] != null)
                    return grid[r, c + 1];

                r--;
                c--;
            }
            return grid[i + 1, j + 1];
        }

        /** Find the nearest point in the N direction */
        private Point NearestNeighbourN(int i, int j, int dist)
        {
            int c1, c2, c3, r;
            if (dist != 0)
            {
                if (grid[i - 1, j] != null)
                    return grid[i - 1, j];

                c1 = j - 1;
                c2 = j;
                c3 = j + 1;
                r = i - 2;

                for (int k = 1; k < dist; k++)
                {
                    if (grid[r, c2] != null)
                        return grid[r, c2];
                    if (grid[r, c1] != null)
                        return grid[r, c1];
                    if (grid[r, c3] != null)
                        return grid[r, c3];

                    r--;
                }
            }
            return grid[i + 1, j];
        }

        /** Find the nearest point in the NE direction */
        private Point NearestNeighbourNE(int i, int j, int dist)
        {
            int r = i - 1;
            int c = j + 1;

            for (int k = 0; k < dist; k++)
            {
                if (grid[r, c] != null)
                    return grid[r, c];
                if (grid[r, c - 1] != null)
                    return grid[r, c - 1];
                if (grid[r + 1, c] != null)
                    return grid[r + 1, c];

                r--;
                c++;
            }
            return grid[i + 1, j - 1];
        }

        /** Find the nearest point in the E direction */
        private Point NearestNeighbourE(int i, int j, int dist)
        {
            int r1, r2, r3, c;
            if (dist != 0)
            {
                if (grid[i, j + 1] != null)
                    return grid[i, j + 1];

                r1 = i - 1;
                r2 = i;
                r3 = i + 1;
                c = j + 2;

                for (int k = 1; k < dist; k++)
                {
                    if (grid[r2, c] != null)
                        return grid[r2, c];
                    if (grid[r1, c] != null)
                        return grid[r1, c];
                    if (grid[r3, c] != null)
                        return grid[r3, c];

                    c++;
                }
            }
            return grid[i, j - 1];
        }

        /** Find the nearest point in the SE direction */
        private Point NearestNeighbourSE(int i, int j, int dist)
        {
            int r = i + 1;
            int c = j + 1;

            for (int k = 0; k < dist; k++)
            {
                if (grid[r, c] != null)
                    return grid[r, c];
                if (grid[r - 1, c] != null)
                    return grid[r - 1, c];
                if (grid[r, c - 1] != null)
                    return grid[r, c - 1];

                r++;
                c++;
            }
            return grid[i - 1, j - 1];
        }

        /** Find the nearest point in the S direction */
        private Point NearestNeighbourS(int i, int j, int dist)
        {
            int c1, c2, c3, r;
            if (dist != 0)
            {
                if (grid[i + 1, j] != null)
                    return grid[i + 1, j];

                c1 = j + 1;
                c2 = j;
                c3 = j - 1;
                r = i + 2;

                for (int k = 1; k < dist; k++)
                {
                    if (grid[r, c2] != null)
                        return grid[r, c2];
                    if (grid[r, c1] != null)
                        return grid[r, c1];
                    if (grid[r, c3] != null)
                        return grid[r, c3];

                    r++;
                }
            }
            return grid[i - 1, j];
        }

        /** Find the nearest point in the SW direction */
        private Point NearestNeighbourSW(int i, int j, int dist)
        {
            int r = i + 1;
            int c = j - 1;

            for (int k = 0; k < dist; k++)
            {
                if (grid[r, c] != null)
                    return grid[r, c];
                if (grid[r, c + 1] != null)
                    return grid[r, c + 1];
                if (grid[r - 1, c] != null)
                    return grid[r - 1, c];

                r++;
                c--;
            }
            return grid[i - 1, j + 1];
        }

        /** Find the nearest point in the W direction */
        private Point NearestNeighbourW(int i, int j, int dist)
        {
            int r1, r2, r3, c;

            if (dist != 0)
            {
                if (grid[i, j - 1] != null)
                    return grid[i, j - 1];

                r1 = i + 1;
                r2 = i;
                r3 = i - 1;
                c = j - 2;

                for (int k = 1; k < dist; k++)
                {
                    if (grid[r2, c] != null)
                        return grid[r2, c];
                    if (grid[r1, c] != null)
                        return grid[r1, c];
                    if (grid[r3, c] != null)
                        return grid[r3, c];

                    c--;
                }
            }
            return grid[i, j + 1];
        }

        /** Every point is assigned to the terrain feature category 
          (flat, opponent_aspect_edge, terrain_edge, uniform_slope, curved_slope) */
        private void ClassifyPoints()
        {
            int maxI = grid.GetLength(0) - 2;
            int maxJ = grid.GetLength(1) - 2;

            for (int i = 2; i < maxI; i++)
            {
                for (int j = 2; j < maxJ; j++)
                {
                    if (grid[i, j] == null)
                        continue;

                    if (grid[i, j].slope < 3) // if slope is less than 3 then point is classified as FLAT
                    {
                        grid[i, j].terFeature = TerrainFeatureE.FLAT;
                    }
                    else
                    {
                        if (grid[i, j].Ha > 0.18 && grid[i, j].Hs > 0.15) // if point lies on the edge of two slopes with the opposite aspect
                        {
                            grid[i, j].terFeature = TerrainFeatureE.OPP_ASPECT;
                        }
                        else
                        {

                            if (grid[i, j].Hs > 0.28) // if point lies on the edge where slope is changing significantly
                            {
                                grid[i, j].terFeature = TerrainFeatureE.TERRAIN_EDGES;
                            }
                            else
                            {
                                if (grid[i, j].slope > 15 && grid[i, j].Hs < 0.15) // Uniform slope - minimum curvature on the slope
                                {
                                    grid[i, j].terFeature = TerrainFeatureE.UNIFORM_SLOPE;
                                }
                                else if (grid[i, j].slope > 15)
                                {
                                    grid[i, j].terFeature = TerrainFeatureE.CURVED_SLOPE; // Curved slope
                                }
                                else
                                {
                                    grid[i, j].terFeature = TerrainFeatureE.NONE;
                                }
                            }
                        }
                    }
                }
            }
        }

        /** Create an 2D array of areas 5x5 if it is possible */
        private void CreateSquareAreas()
        {
            int i, j, k, l;
            int width = grid.GetLength(1) - 4; // 2 border lines dont have calculated all parameters
            int height = grid.GetLength(0) - 4;
            int modX = width % SQUARE_SIZE; // number of rest cells in a row
            int modY = height % SQUARE_SIZE; // number of rest cells in a column
            int squaresX = width / SQUARE_SIZE; // number of square areas in a row
            int squaresY = height / SQUARE_SIZE; // number of square areas in a column
           
            if (modX == 0)
            {
                if (modY == 0)
                {
                    areas = new Area[squaresY, squaresX];
                }
                else
                {
                    areas = new Area[squaresY + 1, squaresX];
                }
            }
            else
            {
                if (modY == 0)
                {
                    areas = new Area[squaresY, squaresX + 1];
                }
                else
                {
                    areas = new Area[squaresY + 1, squaresX + 1];
                }
            }

            int x, y; // left upper corner coordinates (indeces) of the area
            int maxX, maxY; // right bottom corner coordinates of the area
            int pointsCounter = 0; // counter of all points in the area
            int flatCounter = 0;
            int slopeCounter = 0;
            int oppAspectEdgeCounter = 0;
            int edgeCounter = 0;
            int otherPointCounter = 0;
            float sumZ = 0;

            // go through all square areas except border rectangle areas
            for (i = 0; i < squaresY; i++)
            {
                for (j = 0; j < squaresX; j++)
                {
                    x = j * SQUARE_SIZE + 2;
                    y = i * SQUARE_SIZE + 2;
                    maxX = x + SQUARE_SIZE;
                    maxY = y + SQUARE_SIZE;
                    
                    for (k = y; k < maxY; k++)
                    {
                        for (l = x; l < maxX; l++)
                        {
                            if (grid[k, l] != null)
                            {
                                pointsCounter++;
                                sumZ += grid[k, l].z;

                                switch (grid[k, l].terFeature)
                                {
                                    case TerrainFeatureE.FLAT:
                                        flatCounter++;
                                        break;
                                    case TerrainFeatureE.OPP_ASPECT:
                                        oppAspectEdgeCounter++;
                                        break;
                                    case TerrainFeatureE.TERRAIN_EDGES:
                                        edgeCounter++;
                                        break;
                                    case TerrainFeatureE.UNIFORM_SLOPE:
                                        slopeCounter++;
                                        break;
                                    case TerrainFeatureE.CURVED_SLOPE:
                                        slopeCounter++;
                                        break;
                                    default:
                                        otherPointCounter++;
                                        break;
                                }
                            }
                        }
                    }

                    areas[i, j] = new Area(x, y, i, j, SQUARE_SIZE, SQUARE_SIZE, pointsCounter, sumZ, flatCounter, slopeCounter, edgeCounter, oppAspectEdgeCounter, otherPointCounter);
                    pointsCounter = 0;
                    sumZ = 0;
                    flatCounter = 0;
                    slopeCounter = 0;
                    oppAspectEdgeCounter = 0;
                    edgeCounter = 0;
                    otherPointCounter = 0;
                }
            }

            if (modX != 0)
            {
                x = squaresX * SQUARE_SIZE + 2;
                maxX = x + modX;

                // go through all rectangle areas of the last column except the right bottom corner
                for (i = 0; i < squaresY; i++)
                {
                    y = i * SQUARE_SIZE + 2;
                    maxY = y + modY;

                    for (k = y; k < maxY; k++)
                    {
                        for (l = x; l < maxX; l++)
                        {
                            if (grid[k, l] != null)
                            {
                                pointsCounter++;
                                sumZ += grid[k, l].z;

                                switch (grid[k, l].terFeature)
                                {
                                    case TerrainFeatureE.FLAT:
                                        flatCounter++;
                                        break;
                                    case TerrainFeatureE.OPP_ASPECT:
                                        oppAspectEdgeCounter++;
                                        break;
                                    case TerrainFeatureE.TERRAIN_EDGES:
                                        edgeCounter++;
                                        break;
                                    case TerrainFeatureE.UNIFORM_SLOPE:
                                        slopeCounter++;
                                        break;
                                    case TerrainFeatureE.CURVED_SLOPE:
                                        slopeCounter++;
                                        break;
                                    default:
                                        otherPointCounter++;
                                        break;
                                }
                            }
                        }
                    }
                    areas[i, squaresX] = new Area(x, y, i, squaresX, modX, SQUARE_SIZE, pointsCounter, sumZ, flatCounter, slopeCounter, edgeCounter, oppAspectEdgeCounter, otherPointCounter);
                    pointsCounter = 0;
                    sumZ = 0;
                    flatCounter = 0;
                    slopeCounter = 0;
                    oppAspectEdgeCounter = 0;
                    edgeCounter = 0;
                    otherPointCounter = 0;
                }
            }

            if (modY != 0)
            {
                y = squaresY * SQUARE_SIZE + 2;
                maxY = y + modY;
                // go through all rectangle areas of the last row except the right bottom corner
                for (i = 0; i < squaresX; i++)
                {
                    x = i * SQUARE_SIZE + 2;
                    maxX = x + modX;

                    for (k = y; k < maxY; k++)
                    {
                        for (l = x; l < maxX; l++)
                        {
                            if (grid[k, l] != null)
                            {
                                pointsCounter++;
                                sumZ += grid[k, l].z;

                                switch (grid[k, l].terFeature)
                                {
                                    case TerrainFeatureE.FLAT:
                                        flatCounter++;
                                        break;
                                    case TerrainFeatureE.OPP_ASPECT:
                                        oppAspectEdgeCounter++;
                                        break;
                                    case TerrainFeatureE.TERRAIN_EDGES:
                                        edgeCounter++;
                                        break;
                                    case TerrainFeatureE.UNIFORM_SLOPE:
                                        slopeCounter++;
                                        break;
                                    case TerrainFeatureE.CURVED_SLOPE:
                                        slopeCounter++;
                                        break;
                                    default:
                                        otherPointCounter++;
                                        break;
                                }
                            }
                        }
                    }
                    areas[squaresY, i] = new Area(x, y, squaresY, i, SQUARE_SIZE, modY, pointsCounter, sumZ, flatCounter, slopeCounter, edgeCounter, oppAspectEdgeCounter, otherPointCounter);
                    pointsCounter = 0;
                    sumZ = 0;
                    flatCounter = 0;
                    slopeCounter = 0;
                    oppAspectEdgeCounter = 0;
                    edgeCounter = 0;
                    otherPointCounter = 0;
                }

                // add the right bottom rectangle area
                if (modX != 0)
                {
                    x = squaresX * SQUARE_SIZE + 2;
                    y = squaresY * SQUARE_SIZE + 2;
                    maxX = x + modX;
                    maxY = y + modY;

                    for (k = y; k < maxY; k++)
                    {
                        for (l = x; l < maxX; l++)
                        {
                            if (grid[k, l] != null)
                            {
                                pointsCounter++;
                                sumZ += grid[k, l].z;
                                switch(grid[k, l].terFeature)
                                {
                                    case TerrainFeatureE.FLAT:
                                        flatCounter++;
                                        break;
                                    case TerrainFeatureE.OPP_ASPECT:
                                        oppAspectEdgeCounter++;
                                        break;
                                    case TerrainFeatureE.TERRAIN_EDGES:
                                        edgeCounter++;
                                        break;
                                    case TerrainFeatureE.UNIFORM_SLOPE:
                                        slopeCounter++;
                                        break;
                                    case TerrainFeatureE.CURVED_SLOPE:
                                        slopeCounter++;
                                        break;
                                    default:
                                        otherPointCounter++;
                                        break;
                                }
                            }
                        }
                    }
                    areas[squaresY, squaresX] = new Area(x, y, squaresY, squaresX, modX, modY, pointsCounter, sumZ, flatCounter, slopeCounter, edgeCounter, oppAspectEdgeCounter, otherPointCounter);
                }
            }
        }

        private void AnalyzeSquareAreas()
        {
            FindWaterAreas();
            FindEdges();
            FindLocalHighestAndLowestAreas();
        }
        
        /** Classify water areas */
        private void FindWaterAreas()
        {
            int maxI = areas.GetLength(0) - 1;
            int maxJ = areas.GetLength(1) - 1;

            List<Area> waterAreas = new List<Area>();
            List<Area> tempAreas = new List<Area>();
            Queue<Area> exploringAreas = new Queue<Area>();
            
            Area a;
            int counter;

            for (int i = 1; i < maxI; i++)
            {
                for (int j = 1; j < maxJ; j++)
                {
                    if ((areas[i, j].pointsNo == 0) && !areas[i, j].visited) // if area is empty and was not visited yet
                    {
                        
                        a = areas[i, j];
                        a.visited = true;
                        waterAreas.Add(a);
                        exploringAreas.Enqueue(a);

                        // test all neighbouring areas if they are empty
                        do
                        {   a = exploringAreas.Dequeue();
                            tempAreas.Clear();
                            counter = 0;
                            if (IsEmptyArea(a.i - 1, a.j - 1, tempAreas))
                                counter++;
                            if (IsEmptyArea(a.i - 1, a.j, tempAreas))
                                counter++;
                            if (IsEmptyArea(a.i - 1, a.j + 1, tempAreas))
                                counter++;
                            if (IsEmptyArea(a.i, a.j + 1, tempAreas))
                                counter++;
                            if (IsEmptyArea(a.i + 1, a.j + 1, tempAreas))
                                counter++;
                            if (IsEmptyArea(a.i + 1, a.j, tempAreas))
                                counter++;
                            if (IsEmptyArea(a.i + 1, a.j - 1, tempAreas))
                                counter++;
                            if (IsEmptyArea(a.i, a.j - 1, tempAreas))
                                counter++;

                            if (counter > 3) // at least 4 neighbouring areas has to be empty
                            {
                                foreach (Area area in tempAreas)
                                {
                                    waterAreas.Add(area);
                                    exploringAreas.Enqueue(area);
                                }
                            }
                        } while (exploringAreas.Count > 0);

                        if (waterAreas.Count > 49) // if at least 50 areas adjacent to each other
                        {
                            foreach (Area area in waterAreas)
                            {
                                area.landform = LandformsE.WATER;
                            }
                        }
                      
                        waterAreas.Clear();
                    }
                }
            }
        }


        /** Test if the area has no points */
        private bool IsEmptyArea(int i, int j, List<Area> tempAreas)
        {
            if ((i < 0) || (j < 0) || (i == areas.GetLength(0)) || (j == areas.GetLength(1)))
            { 
                return false;
            }
            if (areas[i, j].visited)
            {
                return true;
            }
            if (areas[i, j].pointsNo == 0)
            {
                areas[i, j].visited = true;
                tempAreas.Add(areas[i, j]);
                return true;
            }
            return false;
        }

        /** Find all areas which contains terrain edges and each edge store as a list of points */
        private void FindEdges()
        {
            int maxI = areas.GetLength(0) - 1;
            int maxJ = areas.GetLength(1) - 1;

            List<List<Area>> edges = new List<List<Area>>();
            List<Area> edgeAreas;
            Queue<Area> exploringAreas = new Queue<Area>();

            Area a;

            for (int i = 1; i < maxI; i++)
            {
                for (int j = 1; j < maxJ; j++)
                {
                    if ((areas[i, j].oppAspectPointNo > 1) && !areas[i, j].visited) // if area contains at least 2 points of opp. aspect edge and not visited yet
                    {
                        edgeAreas = new List<Area>();

                        a = areas[i, j];
                        a.visited = true;
                        edgeAreas.Add(a);
                        exploringAreas.Enqueue(a);

                        // test all neighbouring areas if they contains edge
                        do
                        {
                            a = exploringAreas.Dequeue();

                            IsPartOfEdge(a.i - 1, a.j - 1, edgeAreas, exploringAreas);
                            IsPartOfEdge(a.i - 1, a.j, edgeAreas, exploringAreas);
                            IsPartOfEdge(a.i - 1, a.j + 1, edgeAreas, exploringAreas);
                            IsPartOfEdge(a.i, a.j + 1, edgeAreas, exploringAreas);
                            IsPartOfEdge(a.i + 1, a.j + 1, edgeAreas, exploringAreas);
                            IsPartOfEdge(a.i + 1, a.j, edgeAreas, exploringAreas);
                            IsPartOfEdge(a.i + 1, a.j - 1, edgeAreas, exploringAreas);
                            IsPartOfEdge(a.i, a.j - 1, edgeAreas, exploringAreas);
   
                        } while (exploringAreas.Count > 0);

                        AnalyzeEdgeSurrounding(edgeAreas);
                        edges.Add(edgeAreas);
                    }   
                }
            }
        }

        /** Test if area contains edge points */
        private void IsPartOfEdge(int i, int j, List<Area> edgeAreas, Queue<Area> exploringAreas)
        {
            if ((i < 0) || (j < 0) || (i == areas.GetLength(0)) || (j == areas.GetLength(1))) return;

            if (!areas[i, j].visited && areas[i, j].oppAspectPointNo > 0)
            {
                edgeAreas.Add(areas[i, j]);
                exploringAreas.Enqueue(areas[i, j]);
                areas[i, j].visited = true;
            }
        }

        /** Find adjacent slope areas and compare their average z with average z of the edge. 
            Decide if areas are part of a hill or a valley.  (it is not working well) */
        private void AnalyzeEdgeSurrounding(List<Area> edgeAreas)
        {
            float avgEdgeZ, avgSlopeZ;
            float sumEdgeZ = 0;
            float sumSlopeZ = 0;
            Area ar;
            List<Area> slopeAreas = new List<Area>();
            Queue<Area> exploringAreas = new Queue<Area>();

            foreach (Area a in edgeAreas) // go through all points of the edge
            {
                sumEdgeZ += a.avgZ;
                IsSlopeArea(a.i - 1, a.j - 1, slopeAreas, exploringAreas);
                IsSlopeArea(a.i - 1, a.j, slopeAreas, exploringAreas);
                IsSlopeArea(a.i - 1, a.j + 1, slopeAreas, exploringAreas);
                IsSlopeArea(a.i, a.j + 1, slopeAreas, exploringAreas);
                IsSlopeArea(a.i + 1, a.j + 1, slopeAreas, exploringAreas);
                IsSlopeArea(a.i + 1, a.j, slopeAreas, exploringAreas);
                IsSlopeArea(a.i + 1, a.j - 1, slopeAreas, exploringAreas);
                IsSlopeArea(a.i, a.j - 1, slopeAreas, exploringAreas);
            }

            if (slopeAreas.Count > 0)
            {
                do
                {
                    ar = exploringAreas.Dequeue();

                    IsSlopeArea(ar.i - 1, ar.j - 1, slopeAreas, exploringAreas);
                    IsSlopeArea(ar.i - 1, ar.j, slopeAreas, exploringAreas);
                    IsSlopeArea(ar.i - 1, ar.j + 1, slopeAreas, exploringAreas);
                    IsSlopeArea(ar.i, ar.j + 1, slopeAreas, exploringAreas);
                    IsSlopeArea(ar.i + 1, ar.j + 1, slopeAreas, exploringAreas);
                    IsSlopeArea(ar.i + 1, ar.j, slopeAreas, exploringAreas);
                    IsSlopeArea(ar.i + 1, ar.j - 1, slopeAreas, exploringAreas);
                    IsSlopeArea(ar.i, ar.j - 1, slopeAreas, exploringAreas);

                } while (exploringAreas.Count > 0);

                foreach (Area a in slopeAreas)
                {
                    sumSlopeZ += a.avgZ;
                }

                avgEdgeZ = sumEdgeZ / edgeAreas.Count;
                avgSlopeZ = sumSlopeZ / slopeAreas.Count;

                LandformsE landform;
                if (avgEdgeZ < avgSlopeZ) // valley
                {
                    landform = LandformsE.VALLEY;
                }
                else // hill
                {
                    landform = LandformsE.HILL;
                }

                foreach (Area a in slopeAreas)
                {
                    a.landform = landform;
                }
            }
        }

        /** Test if area is slope */
        private void IsSlopeArea(int i , int j, List<Area> slopeAreas, Queue<Area> exploringAreas)
        {
            if ((i < 0) || (j < 0) || (i == areas.GetLength(0)) || (j == areas.GetLength(1))) return;

            if (!areas[i, j].visited && areas[i, j].ratioSlope > 0.8)
            {
                slopeAreas.Add(areas[i, j]);
                exploringAreas.Enqueue(areas[i, j]);
                areas[i, j].visited = true;
            }
        }

        // Find local lowest and highest areas according to the average z coordinate.
        private void FindLocalHighestAndLowestAreas()
        {
            float midAvgZ;
            int maxI = areas.GetLength(0) - 1;
            int maxJ = areas.GetLength(1) - 1;

            for (int i = 1; i < maxI; i++)
            {
                for (int j = 1; j < maxJ; j++)
                {
                    if (areas[i, j].pointsNo < 5)
                    {
                        continue;
                    }
                    midAvgZ = areas[i, j].avgZ;

                    // test if average z value of mid cell is higher than all other neighbours
                    if (midAvgZ > areas[i - 1, j - 1].avgZ)
                    {
                        if (midAvgZ > areas[i - 1, j].avgZ)
                        {
                            if (midAvgZ > areas[i - 1, j + 1].avgZ)
                            {
                                if (midAvgZ > areas[i, j + 1].avgZ)
                                {
                                    if (midAvgZ > areas[i + 1, j + 1].avgZ)
                                    {
                                        if (midAvgZ > areas[i + 1, j].avgZ)
                                        {
                                            if (midAvgZ > areas[i + 1, j - 1].avgZ)
                                            {
                                                if (midAvgZ > areas[i, j - 1].avgZ)
                                                {
                                                    highestAreas.Add(areas[i, j]);
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                    else
                    {   // test if average z value of mid cell is lower than all other neighbours
                        if (midAvgZ < areas[i - 1, j].avgZ)
                        {
                            if (midAvgZ < areas[i - 1, j + 1].avgZ)
                            {
                                if (midAvgZ < areas[i, j + 1].avgZ)
                                {
                                    if (midAvgZ < areas[i + 1, j + 1].avgZ)
                                    {
                                        if (midAvgZ < areas[i + 1, j].avgZ)
                                        {
                                            if (midAvgZ < areas[i + 1, j - 1].avgZ)
                                            {
                                                if (midAvgZ < areas[i, j - 1].avgZ)
                                                {
                                                    lowestAreas.Add(areas[i, j]);
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }

        /** Draw terrain divided into areas and colored according to classified landforms (not finished) */
        public void DrawLandforms()
        {
            InitializeCanvas();
          
            float x, y;
            float width, height;
            Color color;
            SolidBrush brush;
            int maxSquareX = areas.GetLength(1);
            int maxSquareY = areas.GetLength(0);
         
            // go through all square areas (border areas are not supposed to be squared)
            for (int i = 0; i < maxSquareY; i++)
            {
                for (int j = 0; j < maxSquareX; j++)
                {
                    x = areas[i, j].x * ratioX;
                    y = areas[i, j].y * ratioY;
                    width = areas[i, j].width * ratioX;
                    height = areas[i, j].height * ratioY;
                    
                    if (areas[i, j].landform.Equals(LandformsE.WATER))
                    {
                        color = Color.Blue;
                    }
                    else if (areas[i, j].containsEdge)
                    {
                        color = Color.Black;
                    }
                    /*
                    else if (areas[i, j].landform.Equals(LandformsE.VALLEY))
                    {
                        color = Color.Yellow;
                    }
                    else if (areas[i, j].landform.Equals(LandformsE.HILL))
                    {
                        color = Color.Violet;
                    }
                    */
                    else if (areas[i, j].ratioSlope > 0.8)
                    {
                        color = Color.Yellow;
                    }
                    else if (areas[i, j].ratioFlat > 0.8)
                    {
                        color = Color.Green;
                    }
                    else
                    {
                        color = Color.Gray;
                    }
                    
                    brush = new SolidBrush(color);
                    g.FillRectangle(brush, x, y, width, height);
                }
            }
            
            // draw local lowest areas
            color = Color.White;
            brush = new SolidBrush(color);
            foreach(Area a in lowestAreas)
            {
                if (a.ratioSlope > 0.8 || a.oppAspectPointNo > 2)
                {
                    continue;
                }
                x = a.x * ratioX;
                y = a.y * ratioY;
                width = a.width * ratioX;
                height = a.height * ratioY;
                g.FillRectangle(brush, x, y, width, height);
            }
            
            // draw local highest areas
            color = Color.Aqua;
            brush = new SolidBrush(color);
            foreach (Area a in highestAreas)
            {
                if (a.ratioSlope > 0.8 || a.oppAspectPointNo > 2)
                {
                    continue;
                }
                x = a.x * ratioX;
                y = a.y * ratioY;
                width = a.width * ratioX;
                height = a.height * ratioY;
                g.FillRectangle(brush, x, y, width, height);
            }
            
            canvas.Invalidate();
            availableZoom = false;
        }

        /** Draw terrain features */
        public void DrawTerrainFeatures()
        {
            InitializeCanvas();

            int maxI = grid.GetLength(0) - 2;
            int maxJ = grid.GetLength(1) - 2;

            float x, y;
            Color color;
            Color col1, col2, col3, col4, col5, col6;
            SolidBrush brush;

            // initialize colors according to the background color
            if (bgColor.Equals(Color.Black))
            {
                col1 = Color.Red;
                col2 = Color.Aqua;
                col3 = Color.Violet;
                col4 = Color.Yellow;
                col5 = Color.Orange;
                col6 = Color.Gray;
            }
            else
            {
                col1 = Color.Green;
                col2 = Color.Blue;
                col3 = Color.DarkViolet;
                col4 = Color.Red;
                col5 = Color.DarkOrange;
                col6 = Color.DarkGray;
            }

            for (int i = 2; i < maxI; i++)
            {
                for (int j = 2; j < maxJ; j++)
                {
                    if (grid[i, j] == null)
                        continue;

                    x = (float)Math.Round((grid[i, j].x - rd.minX) * ratioX);
                    y = canvas.Height - 1 - (float)Math.Round((grid[i, j].y - rd.minY) * ratioY);

                    switch(grid[i, j].terFeature)
                    {
                        case TerrainFeatureE.FLAT:
                            if (terFeatures[0])
                                color = col1;
                            else
                                color = col6;
                            break;
                        case TerrainFeatureE.OPP_ASPECT:
                            if (terFeatures[1])
                                color = col2;
                            else
                                color = col6;
                            break;
                        case TerrainFeatureE.TERRAIN_EDGES:
                            if (terFeatures[2])
                                color = col3;
                            else
                                color = col6;
                            break;
                        case TerrainFeatureE.UNIFORM_SLOPE:
                            if (terFeatures[3])
                                color = col4;
                            else
                                color = col6;
                            break;
                        case TerrainFeatureE.CURVED_SLOPE:
                            if (terFeatures[4])
                                color = col5;
                            else
                                color = col6;
                            break;
                        default:
                            color = col6;
                            break;
                    }
                    
                    imageBuff[i, j] = color;
                    brush = new SolidBrush(color);
                    g.FillEllipse(brush, x, y, POINT_SIZE, POINT_SIZE);
                }
            }
            canvas.Invalidate();
            SetZoom();
        }


        /* Draw all points and color them according to Z-coordinate */
        public void DrawTerrainZ()
        {
            InitializeCanvas();

            int maxI = grid.GetLength(0) - 1;
            int maxJ = grid.GetLength(1) - 1;

            float x, y;
            float colorCoef = 255 / rd.rangeZ;
            Color color;
            SolidBrush brush;
      
            for (int i = 1; i < maxI; i++)
            {
                for (int j = 1; j < maxJ; j++)
                {
                    if (grid[i, j] == null)
                        continue;

                    x = (float)Math.Round((grid[i, j].x - rd.minX) * ratioX);
                    y = canvas.Height - 1 - (float)Math.Round((grid[i, j].y - rd.minY) * ratioY);

                    color = Color.FromArgb(0, 170, (int)Math.Round((grid[i, j].z - rd.minZ) * colorCoef));
                    
                    imageBuff[i, j] = color;
                    brush = new SolidBrush(color);
                    g.FillEllipse(brush, x, y, POINT_SIZE, POINT_SIZE);
                }
            }
            canvas.Invalidate();
            SetZoom();
        }

        /* Draw all points and color them according to the slope parameter */
        public void DrawTerrainSlope()
        {
            InitializeCanvas();

            int maxI = grid.GetLength(0) - 1;
            int maxJ = grid.GetLength(1) - 1;

            float x, y;
            Color color;
            SolidBrush brush;

            for (int i = 1; i < maxI; i++)
            {
                for (int j = 1; j < maxJ; j++)
                {
                    if (grid[i, j] == null)
                        continue;

                    x = (float)Math.Round((grid[i, j].x - rd.minX) * ratioX);
                    y = canvas.Height - 1 - (float)Math.Round((grid[i, j].y - rd.minY) * ratioY);

                    if (bgColor.Equals(Color.Black))
                    {
                        if (grid[i, j].slope < 5)
                            color = Color.LightGray;
                        else if (grid[i, j].slope >= 5 && grid[i, j].slope < 10)
                            color = Color.Red;
                        else if (grid[i, j].slope >= 10 && grid[i, j].slope < 20)
                            color = Color.Orange;
                        else if (grid[i, j].slope >= 20 && grid[i, j].slope < 30)
                            color = Color.Yellow;
                        else if (grid[i, j].slope >= 30 && grid[i, j].slope < 40)
                            color = Color.Green;
                        else
                            color = Color.Blue;
                    }
                    else
                    {
                        if (grid[i, j].slope < 5)
                            color = Color.Gray;
                        else if (grid[i, j].slope >= 5 && grid[i, j].slope < 10)
                            color = Color.Orange;
                        else if (grid[i, j].slope >= 10 && grid[i, j].slope < 20)
                            color = Color.Red;
                        else if (grid[i, j].slope >= 20 && grid[i, j].slope < 30)
                            color = Color.Green;
                        else if (grid[i, j].slope >= 30 && grid[i, j].slope < 40)
                            color = Color.Blue;
                        else
                            color = Color.DarkViolet;
                    }
                     
                    imageBuff[i, j] = color;
                    brush = new SolidBrush(color);
                    g.FillEllipse(brush, x, y, POINT_SIZE, POINT_SIZE);
                }
            }
            canvas.Invalidate();
            SetZoom();
        }

        /* Draw all points and color them according to aspect paremeter */
        public void DrawTerrainAspect()
        {
            InitializeCanvas();

            int maxI = grid.GetLength(0) - 1;
            int maxJ = grid.GetLength(1) - 1;

            float x, y;
            Color color = Color.Black;
            SolidBrush brush;

            for (int i = 1; i < maxI; i++)
            {
                for (int j = 1; j < maxJ; j++)
                {
                    if (grid[i, j] == null)
                        continue;

                    x = (float)Math.Round((grid[i, j].x - rd.minX) * ratioX);
                    y = canvas.Height - 1 - (float)Math.Round((grid[i, j].y - rd.minY) * ratioY);


                    switch (grid[i, j].aspectDir)
                    {
                        case Dir.N:
                            if (bgColor.Equals(Color.Black))
                                color = Color.White;
                            else
                                color = Color.Black;
                            break;
                        case Dir.NE:
                            color = Color.Red;
                            break;
                        case Dir.E:
                            color = Color.Orange;
                            break;
                        case Dir.SE:
                            if (bgColor.Equals(Color.Black))
                                color = Color.Yellow;
                            else
                                color = Color.Gray;
                            break;
                        case Dir.S:
                            color = Color.Green;
                            break;
                        case Dir.SW:
                            color = Color.Aqua;
                            break;
                        case Dir.W:
                            color = Color.Blue;
                            break;
                        case Dir.NW:
                            color = Color.Violet;
                            break;
                        case Dir.FLAT:
                            color = Color.LightGray;
                            break;
                        case Dir.UNKNOWN:
                            color = Color.Brown;
                            break;
                    }

                    imageBuff[i, j] = color;
                    brush = new SolidBrush(color);
                    g.FillEllipse(brush, x, y, POINT_SIZE, POINT_SIZE);
                }
            }
            canvas.Invalidate();
            SetZoom();
        }

        /* Draw all points and color them according to profile curvature
         * green ... concave
         * red   ... convex
         * white ... no curvature
         */
        public void DrawTerrainProfileCurvature()
        {
            InitializeCanvas();

            int maxI = grid.GetLength(0) - 1;
            int maxJ = grid.GetLength(1) - 1;

            float x, y;
            Color color;
            SolidBrush brush;
            float cur;

            for (int i = 1; i < maxI; i++)
            {
                for (int j = 1; j < maxJ; j++)
                {
                    if (grid[i, j] == null)
                        continue;

                    x = (float)Math.Round((grid[i, j].x - rd.minX) * ratioX);
                    y = canvas.Height - 1 - (float)Math.Round((grid[i, j].y - rd.minY) * ratioY);

                    cur = grid[i, j].profileCurvature;

                    if (cur <= -50) // Convex profile curvature
                    {
                        if (bgColor.Equals(Color.Black))
                            color = Color.Yellow;
                        else
                            color = Color.Red;
                    }
                    else if (cur >= 50) // Concave profile curvature
                        color = Color.Blue;
                    else
                        color = Color.Gray;
                        
                    imageBuff[i, j] = color;
                    brush = new SolidBrush(color);
                    g.FillEllipse(brush, x, y, POINT_SIZE, POINT_SIZE);
                }
            }
            canvas.Invalidate();
            SetZoom();
        }

        /* Draw all points and color them according to plan curvature
         * green ... convex
         * red   ... concave
         * white ... no curvature
         */
        public void DrawTerrainPlanCurvature()
        {
            InitializeCanvas();

            int maxI = grid.GetLength(0) - 1;
            int maxJ = grid.GetLength(1) - 1;

            float x, y;
            Color color;
            SolidBrush brush;

            for (int i = 1; i < maxI; i++)
            {
                for (int j = 1; j < maxJ; j++)
                {
                    if (grid[i, j] == null)
                        continue;

                    x = (float)Math.Round((grid[i, j].x - rd.minX) * ratioX);
                    y = canvas.Height - 1 - (float)Math.Round((grid[i, j].y - rd.minY) * ratioY);

                    double cur = grid[i, j].planCurvature;

                    if (cur <= -30)
                        if (bgColor.Equals(Color.Black))
                            color = Color.Yellow;
                        else
                            color = Color.Red;
                    else if (cur >= 30)
                        color = Color.Blue;
                    else
                        color = Color.Gray;

                    imageBuff[i, j] = color;
                    brush = new SolidBrush(color);
                    g.FillEllipse(brush, x, y, POINT_SIZE, POINT_SIZE);
                }
            }
            canvas.Invalidate();
            SetZoom();
        }

        /** Draw points according to value of overall index homogeneity. */
        public void DrawHImage(double Tall)
        {
            InitializeCanvas();

            int maxI = grid.GetLength(0) - 2;
            int maxJ = grid.GetLength(1) - 2;

            float x, y;
            Color color, col;
            SolidBrush brush;
            double scaledH;

            if (bgColor.Equals(Color.Black))
                col = Color.Yellow;
            else
                col = Color.Blue;

            for (int i = 2; i < maxI; i++)
            {
                for (int j = 2; j < maxJ; j++)
                {
                    if (grid[i, j] == null)
                        continue;

                    x = (float)Math.Round((grid[i, j].x - rd.minX) * ratioX);
                    y = canvas.Height - 1 - (float)Math.Round((grid[i, j].y - rd.minY) * ratioY);

                    scaledH = grid[i, j].homogeneity;
                    
                    /*
                    if (scaledH < 0.1)
                        color = Color.Gray;
                    else if (scaledH >= 0.1 && scaledH < 0.2)
                        color = Color.Red;
                    else if (scaledH >= 0.2 && scaledH < 0.4)
                        color = Color.DarkOrange;
                    else if (scaledH >= 0.4 && scaledH < 0.6)
                        color = Color.Green;
                    else if (scaledH >= 0.6 && scaledH < 0.8)
                        color = Color.Blue;
                    else 
                        color = Color.Black;
                    */
                    
                    if (scaledH > Tall)
                        color = col;
                    else
                    {
                        color = Color.Gray;
                    }
                    

                    imageBuff[i, j] = color;
                    brush = new SolidBrush(color);
                    g.FillEllipse(brush, x, y, POINT_SIZE, POINT_SIZE);
                }
            }
            canvas.Invalidate();
            SetZoom();
        }

        /** Draw points according to value of z-coordinate index homogeneity. */
        public void DrawHimageOfZ(double Tz)
        {
            InitializeCanvas();

            int maxI = grid.GetLength(0) - 2;
            int maxJ = grid.GetLength(1) - 2;

            float x, y;
            Color color, col;
            SolidBrush brush;
            double scaledHz;

            if (bgColor.Equals(Color.Black))
                col = Color.Yellow;
            else
                col = Color.Blue;

            for (int i = 2; i < maxI; i++)
            {
                for (int j = 2; j < maxJ; j++)
                {
                    if (grid[i, j] == null)
                        continue;

                    x = (float)Math.Round((grid[i, j].x - rd.minX) * ratioX);
                    y = canvas.Height - 1 - (float)Math.Round((grid[i, j].y - rd.minY) * ratioY);

                    scaledHz = grid[i, j].Hz;
                   
                    if (scaledHz > Tz)
                        color = col;
                    else
                    {
                        color = Color.Gray;
                    }

                    imageBuff[i, j] = color;
                    brush = new SolidBrush(color);
                    g.FillEllipse(brush, x, y, POINT_SIZE, POINT_SIZE);
                }
            }
            canvas.Invalidate();
            SetZoom();
        }

        /** Draw points according to value of slope index homogeneity. */
        public void DrawHimageOfSlope(double Ts)
        {
            InitializeCanvas();

            int maxI = grid.GetLength(0) - 2;
            int maxJ = grid.GetLength(1) - 2;

            float x, y;
            Color color, col;
            SolidBrush brush;
            double scaledHs;

            if (bgColor.Equals(Color.Black))
                col = Color.Yellow;
            else
                col = Color.Blue;

            for (int i = 2; i < maxI; i++)
            {
                for (int j = 2; j < maxJ; j++)
                {
                    if (grid[i, j] == null)
                        continue;

                    x = (float)Math.Round((grid[i, j].x - rd.minX) * ratioX);
                    y = canvas.Height - 1 - (float)Math.Round((grid[i, j].y - rd.minY) * ratioY);

                    scaledHs = grid[i, j].Hs;
                    
                    if (scaledHs > Ts)
                        color = col;
                    else
                    {
                        color = Color.Gray;
                    }
                    
                    imageBuff[i, j] = color;
                    brush = new SolidBrush(color);
                    g.FillEllipse(brush, x, y, POINT_SIZE, POINT_SIZE);
                }
            }
            canvas.Invalidate();
            SetZoom();
        }

        /** Draw points according to value of aspect index homogeneity. */
        public void DrawHimageOfAspect(double Ta)
        {
            InitializeCanvas();

            int maxI = grid.GetLength(0) - 2;
            int maxJ = grid.GetLength(1) - 2;

            float x, y;
            Color color, col;
            SolidBrush brush;
            double scaledHa;

            if (bgColor.Equals(Color.Black))
                col = Color.Yellow;
            else
                col = Color.Blue;

            for (int i = 2; i < maxI; i++)
            {
                for (int j = 2; j < maxJ; j++)
                {
                    if (grid[i, j] == null)
                        continue;

                    x = (float)Math.Round((grid[i, j].x - rd.minX) * ratioX);
                    y = canvas.Height - 1 - (float)Math.Round((grid[i, j].y - rd.minY) * ratioY);

                    scaledHa = grid[i, j].Ha;
                    
                    if (scaledHa > Ta)
                        color = col;
                    else
                    {
                        color = Color.Gray;
                    }
                    
                    imageBuff[i, j] = color;
                    brush = new SolidBrush(color);
                    g.FillEllipse(brush, x, y, POINT_SIZE, POINT_SIZE);
                }
            }
            canvas.Invalidate();
            SetZoom();
        }

        /** Draw points according to value of profile curvature index homogeneity. */
        public void DrawHimageOfProfileCurvature(double Tpr)
        {
            InitializeCanvas();

            int maxI = grid.GetLength(0) - 2;
            int maxJ = grid.GetLength(1) - 2;

            float x, y;
            Color color, col;
            SolidBrush brush;
            double scaledHpr;

            if (bgColor.Equals(Color.Black))
                col = Color.Yellow;
            else
                col = Color.Blue;

            for (int i = 2; i < maxI; i++)
            {
                for (int j = 2; j < maxJ; j++)
                {
                    if (grid[i, j] == null)
                        continue;

                    x = (float)Math.Round((grid[i, j].x - rd.minX) * ratioX);
                    y = canvas.Height - 1 - (float)Math.Round((grid[i, j].y - rd.minY) * ratioY);

                    scaledHpr = grid[i, j].Hpr;
                    
                    if (scaledHpr > Tpr)
                        color = col;
                    else
                    {
                        color = Color.Gray;
                    }
                    
                    imageBuff[i, j] = color;
                    brush = new SolidBrush(color);
                    g.FillEllipse(brush, x, y, POINT_SIZE, POINT_SIZE);
                }
            }
            canvas.Invalidate();
            SetZoom();
        }

        /** Draw points according to value of plan curvature index homogeneity. */
        public void DrawHimageOfPlanCurvature(double Tpl)
        {
            InitializeCanvas();

            int maxI = grid.GetLength(0) - 2;
            int maxJ = grid.GetLength(1) - 2;
            float x, y;
            Color color, col;
            SolidBrush brush;
            double scaledHpl;

            if (bgColor.Equals(Color.Black))
                col = Color.Yellow;
            else
                col = Color.Blue;

            for (int i = 2; i < maxI; i++)
            {
                for (int j = 2; j < maxJ; j++)
                {
                    if (grid[i, j] == null)
                        continue;

                    x = (float)Math.Round((grid[i, j].x - rd.minX) * ratioX);
                    y = canvas.Height - 1 - (float)Math.Round((grid[i, j].y - rd.minY) * ratioY);

                    scaledHpl = grid[i, j].Hpl;
                    
                    if (scaledHpl > Tpl)
                        color = col;
                    else
                    {
                        color = Color.Gray;
                    }

                    imageBuff[i, j] = color;
                    brush = new SolidBrush(color);
                    g.FillEllipse(brush, x, y, POINT_SIZE, POINT_SIZE);

                }
            }
            canvas.Invalidate();
            SetZoom();
        }

        /** Initialize the canvas and sizes of the painting window */
        private void InitializeCanvas()
        {
            g.Clear(bgColor);
            float ratioA; // ratio of the sides of the area

            // find x and y ratio to fit the area to the canvas
            if (rd.rangeY > rd.rangeX)
            {
                if (rd.rangeY > canvas.Height)
                {
                    ratioY = canvas.Height / rd.rangeY;
                    ratioA = rd.rangeX / rd.rangeY;
                    ratioX = canvas.Height * ratioA / rd.rangeX;
                }
                else // area is smaller than canvas
                {
                    ratioY = 1;
                    ratioX = 1;
                }
            }
            else // area is wider than higher
            {
                if (rd.rangeX > canvas.Width)
                {
                    ratioX = canvas.Width / rd.rangeX;
                    ratioA = rd.rangeY / rd.rangeX;
                    ratioY = canvas.Width * ratioA / rd.rangeY;
                }
                else 
                {
                    ratioY = 1;
                    ratioX = 1;
                }
                
            }
        }

        /** if area is larger than the size of the canvas, then set the zoom available */
        private void SetZoom()
        {   
            if (rd.rangeX > canvas.Width || rd.rangeY > canvas.Height)
                availableZoom = true;
            else
                availableZoom = false;
        }

        /** Draw a zoomed area */
        public void DrawParticularArea(int r, int c)
        {
            g.Clear(bgColor);
  
            row = r;
            column = c;
            int maxR = r + canvas.Height - 1;
            int maxC = c + canvas.Width - 1;
            for (int i = r, k = 0; i < maxR; i++, k++)
            {
                for (int j = c, l = 0; j < maxC; j++, l++)
                {
                    if (imageBuff[i, j] != null)
                    {
                        g.FillEllipse(new SolidBrush(imageBuff[i, j]), l, k, 2, 2);
                    }
                }
            }

            canvas.Invalidate();
            availableZoom = false;
            isActiveZoom = true;
        }

        /** Shift zoomed area in a specific direction */
        public void ShiftArea(int dir)
        {
            int off = 100;
            int maxC = imageBuff.GetLength(1) - canvas.Width;
            int maxR = imageBuff.GetLength(0) - canvas.Height;

            // Shift up
            if (dir == 0)
            {
                if (row == 0)
                    return;
                else if (row < off)
                    DrawParticularArea(0, column);
                else
                    DrawParticularArea(row - off, column);
            }
            // Shift right
            else if (dir == 1)
            {
                if (column == maxC)
                    return;
                else if (column > (maxC - off))
                    DrawParticularArea(row, maxC);
                else
                    DrawParticularArea(row, column + off);
            }
            // Shift down
            else if (dir == 2)
            {
                if (row == maxR)
                    return;
                else if (row > (maxR - off))
                    DrawParticularArea(maxR, column);
                else
                    DrawParticularArea(row + off, column);
            }
            // Shift left
            else
            {
                if (column == 0)
                    return;
                else if (column < off)
                    DrawParticularArea(row, 0);
                else
                    DrawParticularArea(row, column - off);
            }
            
        }

        /** Set the background color of the canvas */
        public void setBackground(bool isBlack)
        {
            if (isBlack)
            {
                bgColor = Color.Black;
            }
            else
            {
                bgColor = Color.White;
            }
        }
        
        /** Return the height of the grid */
        public int GetGridHeight()
        {
            return grid.GetLength(0);
        }
        
        /** Return the width of the grid */
        public int GetGridWidth()
        {
            return grid.GetLength(1);
        }
        
        /** Set the parameter for the segmentation */
        public void SetParameter(int index, bool value)
        {
            pars[index] = value;
        }

        /** Set the terrain feature for drawing */
        public void SetTerrainFeature(int index, bool value)
        {
           terFeatures[index] = value;
        }

    }
}
