﻿using System;
using System.Drawing;
using System.Windows.Forms;

namespace LandformsClassifier
{
    public partial class Window : Form
    {
        private Graphics g;
        private PictureBox can; // canvas
        private Bitmap bitmap; // bitmap of the canvas
        private Terrain terrain; // instance representing all terrain

        OpenFileDialog openFile = new OpenFileDialog();

        public Window()
        {
            // initialization
            InitializeComponent();

            can = canvas;
            bitmap = new Bitmap(can.Width, can.Height);
            can.Image = bitmap;
            g = Graphics.FromImage(bitmap);
            g.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.AntiAlias;

            terrain = new Terrain(g, can);
            initializeCheckList();
            can.Focus();
        }

        /** Initialize checklist of the parameters used for segmentation */
        private void initializeCheckList()
        {
            clParameters.SetItemChecked(0, false);
            clParameters.SetItemChecked(1, true);
            clParameters.SetItemChecked(2, true);
            clParameters.SetItemChecked(3, true);
            clParameters.SetItemChecked(4, true);

            clTerrainFeatures.SetItemChecked(0, true);
            clTerrainFeatures.SetItemChecked(1, true);
            clTerrainFeatures.SetItemChecked(2, true);
            clTerrainFeatures.SetItemChecked(3, true);
            clTerrainFeatures.SetItemChecked(4, true);
        }

        /** Handle clicked on Z-coordinate menu option */
        private void zcoordinateMenuItem_Click(object sender, EventArgs e)
        {
            terrain.DrawTerrainZ();
        }

        /** Handle clicked on Slope menu option */
        private void slopeMenuItem_Click(object sender, EventArgs e)
        {
            terrain.DrawTerrainSlope();
        }

        /** Handle clicked on Aspect menu option */
        private void aspectMenuItem_Click(object sender, EventArgs e)
        {
            terrain.DrawTerrainAspect();
        }

        /** Handle clicked on Profile Curvature menu option */
        private void profileCurvatureMenuItem_Click(object sender, EventArgs e)
        {
            terrain.DrawTerrainProfileCurvature();
        }

        /** Handle clicked on Plan Curvature menu option */
        private void planCurvatureMenuItem_Click(object sender, EventArgs e)
        {
            terrain.DrawTerrainPlanCurvature();
        }

        /** Handle clicked on Hz-image menu option */
        private void hzImageMenuItem_Click(object sender, EventArgs e)
        {
            terrain.DrawHimageOfZ(ParseThreshold(tresholdZtb));
        }

        /** Handle clicked on Hs-image menu option */
        private void hsImageMenuItem_Click(object sender, EventArgs e)
        {
             terrain.DrawHimageOfSlope(ParseThreshold(tresholdStb));
        }

        /** Handle clicked on Ha-image menu option */
        private void haImageMenuItem_Click(object sender, EventArgs e)
        {
             terrain.DrawHimageOfAspect(ParseThreshold(tresholdAtb));
        }

        /** Handle clicked on Hpr-image menu option */
        private void hprImageMenuItem_Click(object sender, EventArgs e)
        {
             terrain.DrawHimageOfProfileCurvature(ParseThreshold(tresholdPrtb));
        }

        /** Handle clicked on Hpl-image menu option */
        private void hplImageMenuItem_Click(object sender, EventArgs e)
        {
             terrain.DrawHimageOfPlanCurvature(ParseThreshold(tresholdPltb));
        }

        /** Handle clicked on H-image menu option */
        private void himageMenuItem_Click(object sender, EventArgs e)
        {
            terrain.ComputeOverallH(ParseThreshold(tresholdZtb), ParseThreshold(tresholdStb), ParseThreshold(tresholdAtb),
                ParseThreshold(tresholdPrtb), ParseThreshold(tresholdPltb));
            terrain.DrawHImage(ParseThreshold(tresholdHtb));
        }

        /** Parse the threshold value from the textbox. */ 
        private double ParseThreshold(TextBox tresholdTB)
        {
            double t;

            try
            {
                t = Double.Parse(tresholdTB.Text);
            }
            catch (Exception ex)
            {
                t = 0;
                tresholdTB.Text = "0";
            }

            if (t < 0)
            {
                t = 0;
                tresholdTB.Text = "0";
            }
            return t;
        }

        /** Handle clicked on Terrain Features menu option */
        private void tarrainFeaturesMenuItem_Click(object sender, EventArgs e)
        {
            terrain.DrawTerrainFeatures();
        }

        /** Handle clicked on Landforms menu option */
        private void landformsMenuItem_Click(object sender, EventArgs e)
        {
            terrain.DrawLandforms();
        }

        /** Handle change of the used parameters for segmentation */
        private void clParameters_ItemCheck(object sender, ItemCheckEventArgs e)
        {
            if (clParameters.GetItemCheckState(e.Index) == CheckState.Checked)
            {
                terrain.SetParameter(e.Index, false);
            }
            else
            {
                terrain.SetParameter(e.Index, true);
            }
            can.Focus();
        }

        /** Handle double clicked on the canvas */
        private void canvas_MouseDoubleClick(object sender, MouseEventArgs e)
        {
            if (!terrain.availableZoom) return;

            if (e.Button == MouseButtons.Left)
            {
                int y = can.Height - 1 - e.Y;

                if (can.Width > terrain.GetGridWidth()) // if area width is smaller than the canvas width 
                {
                    if (e.X > terrain.GetGridWidth())
                    {
                        return;
                    }
                }
                if (can.Height > terrain.GetGridHeight()) // if area height is smaller than the canvas height 
                {
                    if (y > terrain.GetGridHeight())
                    {
                        return;
                    }
                }

                int r = (int)((terrain.GetGridHeight() / (double)can.Height) * e.Y);
                int c = (int)((terrain.GetGridWidth() / (double)can.Width) * e.X);

                int maxR = terrain.GetGridHeight() - canvas.Height;
                int maxC = terrain.GetGridWidth() - canvas.Width;


                if (r > maxR)
                    r = maxR;
                if (c > maxC)
                    c = maxC;

                terrain.DrawParticularArea(r, c);
            }
            can.Focus();
        }

        /** Handle pressed arrow keys for draw a shifted area of the terrain */
        private void Window_KeyDown(object sender, KeyEventArgs e)
        {
            if (!terrain.isActiveZoom) return;

            switch(e.KeyCode)
            {
                case Keys.Up:
                    terrain.ShiftArea(0);
                    break;
                case Keys.Right:
                    terrain.ShiftArea(1);
                    break;
                case Keys.Down:
                    terrain.ShiftArea(2);
                    break;
                case Keys.Left:
                    terrain.ShiftArea(3);
                    break;
            }
        }

        /** Handle change of the background color of the canvas */
        private void radioBtnBlack_CheckedChanged(object sender, EventArgs e)
        {
            if (radioBtnBlack.Checked)
            {
                terrain.setBackground(true);
            }
            else
            {
                terrain.setBackground(false);
            }
            can.Focus();
        }

        /** Handle change of the drawing points with a specific terrain features */
        private void clTerrainFeatures_ItemCheck(object sender, ItemCheckEventArgs e)
        {
            if (clTerrainFeatures.GetItemCheckState(e.Index) == CheckState.Checked)
            {
                terrain.SetTerrainFeature(e.Index, false);
            }
            else
            {
                terrain.SetTerrainFeature(e.Index, true);
            }
            can.Focus();
        }

        /** Handle clicked on Open menu option to load a file from file explorer */
        private void openMenuItem_Click(object sender, EventArgs e)
        {
            openFile.Filter = "Files (*.xyz; *.dat)|*.xyz; *.dat";
            openFile.RestoreDirectory = true;

            if (DialogResult.OK == openFile.ShowDialog())
            {
                if (string.IsNullOrEmpty(openFile.FileName))
                    return;
                else
                {
                    terrain.LoadFile(openFile.FileName);
                    viewMenuItem.Enabled = true;
                }
            }
        }
    }
}
